package dev.fyloz.colorrecipesexplorer.databasemanager

import java.io.File
import java.io.FileInputStream
import java.util.*

/** The latest version of the database supported by the utility. The [DatabaseUpdaterProperties] target version cannot be higher than this. */
internal const val LATEST_DATABASE_VERSION = 3

/** The key of the target version property */
internal const val PROPERTY_TARGET_VERSION = "database.target-version"

/** The key of the database server's url property */
internal const val PROPERTY_URL = "database.url"

/** The key of the database's name property */
internal const val PROPERTY_DB_NAME = "database.name"

/** The key of the database's username property */
internal const val PROPERTY_USERNAME = "database.username"

/** The key of the database's password property */
internal const val PROPERTY_PASSWORD = "database.password"

/** The regular expression to validate a database url */
internal val DATABASE_URL_REGEX =
    Regex("^jdbc:\\w+://(([a-zA-Z0-9]|[a-zA-Z0-9][a-zA-Z0-9\\-]*[a-zA-Z0-9])\\.)*([A-Za-z0-9]|[A-Za-z0-9][A-Za-z0-9\\-]*[A-Za-z0-9])(:\\d+)?/\$")

/** The properties used by the utility */
data class DatabaseUpdaterProperties(
    /** The target version of the database */
    val targetVersion: Int,
    /** The url of the database server */
    val url: String,
    /** The name of the database */
    val dbName: String,
    /** The username of the database's user */
    val username: String,
    /** The password of the database's user */
    val password: String
) {
    init {
        if (targetVersion < 0 || targetVersion > LATEST_DATABASE_VERSION)
            throw InvalidDatabaseUpdaterPropertiesException.InvalidTargetVersion(targetVersion)

        if (!url.matches(DATABASE_URL_REGEX))
            throw InvalidDatabaseUpdaterPropertiesException.InvalidUrl(url)
    }
}

/** DSL for creating an instance of [DatabaseUpdaterProperties]. */
fun databaseUpdaterProperties(
    targetVersion: Int = 0,
    url: String = "jdbc:driver://host:1234/",
    dbName: String = "database",
    username: String = "user",
    password: String = "pass",
    op: DatabaseUpdaterProperties.() -> Unit = {}
) = DatabaseUpdaterProperties(targetVersion, url, dbName, username, password).apply(op)

/** DSL for creating an instance of [DatabaseUpdaterProperties] from the given [properties]. */
fun databaseUpdaterProperties(
    properties: Properties,
    op: DatabaseUpdaterProperties.() -> Unit = {}
) = with(properties) {
    /** Gets a [property] from the given [properties]. Calls exitMissingProperty() when the value is null. */
    fun property(property: String): String =
        getProperty(property) ?: throw InvalidDatabaseUpdaterPropertiesException.MissingProperty(property)

    databaseUpdaterProperties(
        targetVersion = with(property(PROPERTY_TARGET_VERSION)) {
            toIntOrNull() ?: throw InvalidDatabaseUpdaterPropertiesException.TargetVersionFormat(this)
        },
        url = property(PROPERTY_URL),
        dbName = property(PROPERTY_DB_NAME),
        username = property(PROPERTY_USERNAME),
        password = property(PROPERTY_PASSWORD),
        op
    )
}

/** DSL for creating an instance of [DatabaseUpdaterProperties] from the given [file]. */
fun databaseUpdaterProperties(
    file: File,
    op: DatabaseUpdaterProperties.() -> Unit = {}
) = databaseUpdaterProperties(
    properties = Properties().apply {
        if (!file.exists() || !file.isFile)
            throw InvalidDatabaseUpdaterPropertiesException.PropertiesNotFound(file.absolutePath)

        load(FileInputStream(file))
    },
    op
)

/** An exception representing an invalid value in a [DatabaseUpdaterProperties] instance. */
internal sealed class InvalidDatabaseUpdaterPropertiesException(message: String) : Exception(message) {
    /** Thrown when the properties file does not exists. */
    class PropertiesNotFound(val path: String) :
        InvalidDatabaseUpdaterPropertiesException("Could not find properties file at '$path'")

    /** Thrown when the target version property cannot be represented as an [Int]. */
    class TargetVersionFormat(val value: String) :
        InvalidDatabaseUpdaterPropertiesException("Invalid database target version format for value '$value'")

    /** Thrown when the target version property is not in the range {0, [LATEST_DATABASE_VERSION]}. */
    class InvalidTargetVersion(val targetVersion: Int) :
        InvalidDatabaseUpdaterPropertiesException("Invalid database target version '$targetVersion'. The database version must be between 0 and $LATEST_DATABASE_VERSION")

    /** Thrown when the url property does not matches the pattern 'jdbc:{driver}://{host}{:port?}/'. */
    class InvalidUrl(val url: String) :
        InvalidDatabaseUpdaterPropertiesException("Invalid database url '$url'. The proper format is: jdbc:{driver}://{host}{:port?}/")

    /** Thrown when the given [property] is not defined. */
    class MissingProperty(val property: String) :
        InvalidDatabaseUpdaterPropertiesException("The property $property is not defined")
}
